#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <windows.h>
#include <SDL/SDL.h>
#include <fmod.h>
#include <fmod_errors.h>
#include "spectrum.h"
#define	ENABLEREVERB	TRUE
#define RECORDRATE      44100
#define RECORDLEN       (RECORDRATE * 5)
#define RECORD_DELAY_MS       25
#define RECORD_DELAY_SAMPLES  (RECORDRATE * RECORD_DELAY_MS / 1000)
#define SPEC_MON 4      // Which chunk of the spectrum should we monitor 
#define AVG_BUFFER 200  // How many samples should we consider for energy average
#define SPEC_WIDTH 512  // 512 channels in spectrum

int main(int argc, char *argv[]) {
	SDL_Surface *screen;
	SDL_Event event;
	Uint32 white;
	Uint32 black;
	Uint32 yellow;
  SDL_Color colors[256];
	FSOUND_SAMPLE *samp1;
	int channel, originalfreq;
	int w = 512;
	int h = 240;
  int i = 0;
	int colorCount = 0;
	int overCount = 0;
	int xi = 0;
  int yi = 0;
	int desired_bpp = 8;
	Uint32 video_flags = 0;
	int done = 0;
	int energy[SPEC_WIDTH];
	int energyHist[SPEC_MON][AVG_BUFFER];
	int spectrumEnergy;
	int spectrumEnergyHist[SPEC_WIDTH];
	int spectrumAverage = 0;
  int averageBeat[SPEC_MON];
	int beatHist[SPEC_WIDTH];
	int count = 0;
	int spec_count = 0;
	video_flags |= SDL_DOUBLEBUF;
	video_flags |= SDL_HWSURFACE;
	video_flags |= SDL_HWPALETTE;
	//video_flags |= SDL_FULLSCREEN;

  for(i=0;i<256;i++) {
    colors[i].r=i*i; //i*i*3;
    colors[i].g=i*i;//+(i/2);
    colors[i].b=i*i;//+(i/2);
	}

	for(i=0;i<w;i++) { beatHist[i] = 0; }

	if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
		fprintf(stderr,
			"Couldn't initialize SDL: %s\n", SDL_GetError());
		exit(1);
	}
	atexit(SDL_Quit);
	screen = SDL_SetVideoMode(w, h, desired_bpp, video_flags);
	if ( screen == NULL ) {
		fprintf(stderr, "Couldn't set %dx%dx%d video mode: %s\n",
					w, h, desired_bpp, SDL_GetError());
		exit(1);
	}

  // FSOUND INIT ------------------------------------

	if (FSOUND_GetVersion() < FMOD_VERSION) {
		printf("Error : You are using the wrong DLL version!  You should be using FMOD %.02f\n", FMOD_VERSION);
		return 0;
	}
  FSOUND_SetOutput(FSOUND_OUTPUT_DSOUND);
  FSOUND_SetDriver(0);
  FSOUND_SetMixer(FSOUND_MIXER_QUALITY_AUTODETECT);
	if (!FSOUND_Init(44100, 32, FSOUND_INIT_ACCURATEVULEVELS)) {
		printf("Error!\n");
		printf("%s\n", FMOD_ErrorString(FSOUND_GetError()));
		return 0;
	}
	for (i=0; i < FSOUND_Record_GetNumDrivers(); i++) {
		printf("%d - %s\n", i+1, FSOUND_Record_GetDriverName(i));	// print driver names 
	}
	if (!FSOUND_Record_SetDriver(0)) { // Select input sound card (0 = default) 
		printf("Error!\n");
		printf("%s\n", FMOD_ErrorString(FSOUND_GetError()));
		FSOUND_Close();
		return 0;
	}
	printf("FSOUND Output Method : ");
	switch (FSOUND_GetOutput()) {
        case FSOUND_OUTPUT_NOSOUND:    printf("FSOUND_OUTPUT_NOSOUND\n"); break;
        case FSOUND_OUTPUT_WINMM:      printf("FSOUND_OUTPUT_WINMM\n"); break;
        case FSOUND_OUTPUT_DSOUND:     printf("FSOUND_OUTPUT_DSOUND\n"); break;
        case FSOUND_OUTPUT_OSS:        printf("FSOUND_OUTPUT_OSS\n"); break;
        case FSOUND_OUTPUT_ESD:        printf("FSOUND_OUTPUT_ESD\n"); break;
        case FSOUND_OUTPUT_ALSA:       printf("FSOUND_OUTPUT_ALSA\n"); break;       
	};
	printf("FSOUND Mixer         : ");
	switch (FSOUND_GetMixer()) {
		case FSOUND_MIXER_BLENDMODE:	printf("FSOUND_MIXER_BLENDMODE\n"); break;
		case FSOUND_MIXER_MMXP5:		printf("FSOUND_MIXER_MMXP5\n"); break;
		case FSOUND_MIXER_MMXP6:		printf("FSOUND_MIXER_MMXP6\n"); break;
		case FSOUND_MIXER_QUALITY_FPU:	printf("FSOUND_MIXER_QUALITY_FPU\n"); break;
		case FSOUND_MIXER_QUALITY_MMXP5:printf("FSOUND_MIXER_QUALITY_MMXP5\n"); break;
		case FSOUND_MIXER_QUALITY_MMXP6:printf("FSOUND_MIXER_QUALITY_MMXP6\n"); break;
	};
	printf("FSOUND Driver        : %s\n", FSOUND_GetDriverName(FSOUND_GetDriver()));
	printf("FSOUND Record Driver : %s\n", FSOUND_Record_GetDriverName(FSOUND_Record_GetDriver()));
    if (FSOUND_GetOutput() == FSOUND_OUTPUT_OSS) {
	    samp1 = FSOUND_Sample_Alloc(FSOUND_UNMANAGED, RECORDLEN, FSOUND_MONO | FSOUND_8BITS | FSOUND_UNSIGNED, RECORDRATE, 255, 128, 255);
    } else {
        samp1 = FSOUND_Sample_Alloc(FSOUND_UNMANAGED, RECORDLEN, FSOUND_STEREO | FSOUND_16BITS , RECORDRATE, 255, 128, 255);
    }
    FSOUND_DSP_SetActive(FSOUND_DSP_GetFFTUnit(), TRUE);
		FSOUND_Sample_SetMode(samp1, FSOUND_LOOP_NORMAL);	

	if (!FSOUND_Record_StartSample(samp1, TRUE)) {	// start recording
		printf("Error!\n");
		printf("%s\n", FMOD_ErrorString(FSOUND_GetError()));

		FSOUND_Close();
		return 0;
	}
  
  while (!FSOUND_Record_GetPosition()) { // Delay until recording is synced in buffer
    SDL_Delay(1);
  }	
	channel = FSOUND_PlaySound(FSOUND_FREE, samp1);
	originalfreq = FSOUND_GetFrequency(channel);
	// END FSOUND INIT -----------------------------------

	white = SDL_MapRGB(screen->format, 0xff, 0xff, 0xff);
	black = SDL_MapRGB(screen->format, 0x00, 0x00, 0x00);
  yellow = SDL_MapRGB(screen->format, 0xff, 0xff, 0x00);



	SDL_UpdateRect(screen, 0, 0, 0, 0);

	// SDL / FMOD Update & Display Loop ----------------------

	while(SDL_PollEvent(&event)) {  
		switch(event.type) { 
			case SDL_KEYDOWN:          
				printf("Oh! Key press\n");
				break;
			default:
				continue;
		}
	}

	while(done == 0) {
		SDL_PumpEvents();
		Uint8 *keystate = SDL_GetKeyState(NULL);
		if ( keystate[SDLK_RETURN] ) {
			done = 1;
			exit(0);	
		}

    if ( SDL_MUSTLOCK(screen) ) {
        if ( SDL_LockSurface(screen) < 0 ) {
            fprintf(stderr, "Can't lock screen: %s\n", SDL_GetError());
            return 0;
        }
    }


    // FMOD Loop ---------------------------------------



		int playpos, recordpos, diff;
		static int oldrecordpos = 0, oldplaypos = 0;

		playpos = FSOUND_GetCurrentPosition(channel);
		recordpos = FSOUND_Record_GetPosition();

		if (playpos > oldplaypos && recordpos > oldrecordpos)	
		{
			diff = playpos - recordpos;
			if (diff > -RECORD_DELAY_SAMPLES)  {
				FSOUND_SetFrequency(channel, originalfreq - 1000);	// slow it down 
			} else if (diff < -(RECORD_DELAY_SAMPLES * 2)) {
				FSOUND_SetFrequency(channel, originalfreq + 1000);	// speed it up 
			} else {
				FSOUND_SetFrequency(channel, originalfreq);	
      }
		}

		oldplaypos = playpos;
		oldrecordpos = recordpos;

		char beat[5] = "";
 		float *spectrum = FSOUND_DSP_GetSpectrum();

    SDL_FillRect(screen, NULL, black);
		for(xi = 0; xi < w; xi++) {
			
      energy[xi] = (h-((int)(spectrum[((int)(xi/2))]*(float)h)));

		    for(yi = h; yi > 1; yi--) {
			    if(yi < energy[xi] ) {
			      putpixel(screen, xi, yi, white);
			      putpixel(screen, xi+1, yi, white);
					}
				}

				energyHist[0][count] = h-energy[4];
				energyHist[1][count] = h-energy[5];
				energyHist[2][count] = h-energy[4];
				energyHist[3][count] = h-energy[5];
				
				xi++;
				energy[xi] = (h-((int)(spectrum[((int)(xi/2))]*(float)h)));
      
		}

    // get average beat energy, determine beat
		if(count) {
      for(int avgHist = 0; avgHist < SPEC_MON; avgHist++) {

  			if(overCount == 0) {
		      for(int hist = 0; hist < (count-1); hist++) {
            averageBeat[avgHist] += energyHist[avgHist][hist];
					}
			    averageBeat[avgHist] /= count;

				}  else {
			    for(int hist = 0; hist < AVG_BUFFER; hist++) {
            averageBeat[avgHist] += energyHist[avgHist][hist];
					}
			    averageBeat[avgHist] /= AVG_BUFFER;
				}        

			}
		}

    spectrumEnergy = 0;
    // get total energy from the entire spectrum
		for(i = 0; i < SPEC_WIDTH; i++) {
      spectrumEnergy += (h-energy[i]);
		}
		spectrumEnergyHist[spec_count] = spectrumEnergy;
		for(i = 0; i < SPEC_WIDTH; i++) {
      spectrumAverage += spectrumEnergyHist[i];
		}
		spectrumAverage /= SPEC_WIDTH;

    int success = 0;
    for(int avgHist = 0; avgHist < SPEC_MON; avgHist++) {
			int beatPeak = (averageBeat[avgHist] + (int)((h-averageBeat[avgHist])*0.6));
			if(beatPeak >= h) { beatPeak = (h-1); }
			if( energyHist[avgHist][count] > beatPeak) { // && spectrumEnergy > spectrumAverage ) {
        success++;
			}
		}
		if(success >= SPEC_MON) { 
			strcpy(beat,"BEAT"); 
		  beatHist[spec_count] = 1;
      //SDL_FillRect(screen, NULL, black);		
		} else {
      beatHist[spec_count] = 0;
		} 

    for(i=0;i<SPEC_WIDTH;i++) {
      if(beatHist[i]) {
				for(int k = 0; k < 4; k++) {
          putpixel(screen, i, k, black);
        }
			} else {
				for(int k = 0; k < 4; k++) {
          putpixel(screen, i, k, white);
        }
			}
      putpixel(screen, i, 5, black);
		}

		printf("energy[%3d|%3d|%3d|%3d|%3d|%3d|%3d] av[%3d] sav[%5d] ?%4s?\r", energy[2], energy[3], energy[4], energy[5], energy[6], energy[7], energy[8], averageBeat[2], spectrumEnergy, beat);


		// -------------------------------------------------
		
    if ( SDL_MUSTLOCK(screen) ) {
        SDL_UnlockSurface(screen);
    }
    /* Update just the part of the display that we've changed */
    SDL_UpdateRect(screen, 0, 0, 0, 0);
  
    count++; spec_count++;
		if(count >= AVG_BUFFER) { count = 0; overCount++; }
		if(spec_count >= SPEC_WIDTH) { spec_count = 0; }
	}

  FSOUND_Close();
  return(0);
}
